package answer

import (
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"log/slog"
	"net/url"
	"strings"

	"github.com/tmc/langchaingo/agents"

	"github.com/tmc/langchaingo/llms"

	"gitea.stevedudenhoeffer.com/steve/answer/pkg/cache"
	"gitea.stevedudenhoeffer.com/steve/answer/pkg/extractor"
	"gitea.stevedudenhoeffer.com/steve/answer/pkg/search"
	gollm "gitea.stevedudenhoeffer.com/steve/go-llm"
)

var ErrMaxTries = errors.New("maximum number of pages tried reached")
var ErrMaxAnswers = errors.New("maximum number of answers parsed reached")
var ErrTooManyArguments = errors.New("too many arguments")

type Question struct {
	// Question is the question to answer
	Question string

	Model llms.Model

	Search search.Search

	Cache cache.Cache
}

// Answers is a list of answers to a question
type Answers []string

type Options struct {
	// MaxSearches is the maximum possible number of searches to execute for this question.  If this is set to 5, the function could
	// search up to 5 possible times to find an answer.
	MaxSearches int

	// MaxThinks is the maximum number of times to think about a question.  A "Think" is different than a search in that
	// the LLM just breaks the question down into smaller parts and tries to answer them.  This is useful for complex
	// questions that are hard to answer since LLMs are better at answering smaller questions.
	MaxThinks int

	// MaxTries is the absolute maximum number of pages to try to get an answer from.  For instance, if MaxSearches is 5 and
	// 5 pages are tried and no answers are found, the function will return ErrMaxTries.
	MaxTries int

	// OnNewFunction is a callback that, if non-nil, will be called when a new function is called by the LLM.
	// The "answer" and "no_answer" functions are not included in this callback.
	// Return an error to stop the function from being called.
	OnNewFunction func(ctx context.Context, funcName string, question string, parameter string) error
}

var DefaultOptions = Options{
	MaxSearches: 5,
	MaxThinks:   10,
	MaxTries:    5,
}

type Result struct {
	Result string
	Error  error
}

func fanExecuteToolCalls(ctx context.Context, toolBox *gollm.ToolBox, calls []gollm.ToolCall) []Result {
	var results []Result
	var resultsOutput = make(chan Result, len(calls))

	fnCall := func(call gollm.ToolCall) Result {
		str, err := toolBox.Execute(ctx, call)
		if err != nil {
			return Result{
				Error: err,
			}
		}

		return Result{
			Result: str,
		}
	}

	for _, call := range calls {
		go func(call gollm.ToolCall) {
			resultsOutput <- fnCall(call)
		}(call)
	}

	for i := 0; i < len(calls); i++ {
		result := <-resultsOutput
		results = append(results, result)
	}

	close(resultsOutput)

	return results
}

type article struct {
	URL   string
	Title string
	Body  string
}

func extractArticle(ctx context.Context, c cache.Cache, u *url.URL) (res article, err error) {
	defer func() {
		e := recover()

		if e != nil {
			if e, ok := e.(error); ok {
				err = fmt.Errorf("panic: %w", e)
			} else {
				err = fmt.Errorf("panic: %v", e)
			}
		}
	}()

	extractors := extractor.MultiExtractor(
		extractor.CacheExtractor{
			Cache:     c,
			Tag:       "goose",
			Extractor: extractor.GooseExtractor{},
		},
		extractor.CacheExtractor{
			Cache:     c,
			Tag:       "playwright",
			Extractor: extractor.PlaywrightExtractor{},
		},
	)

	a, err := extractors.Extract(ctx, u.String())

	if err != nil {
		return article{
			URL:   "",
			Title: "",
			Body:  "",
		}, err
	}

	return article{
		URL:   a.URL,
		Title: a.Title,
		Body:  a.Body,
	}, nil
}

func doesTextAnswerQuestion(ctx context.Context, q Question, text string) (string, error) {
	var availableTools = []llms.Tool{
		{
			Type: "function",
			Function: &llms.FunctionDefinition{
				Name:        "answer",
				Description: "The answer from the given text that answers the question.",
				Parameters: map[string]any{
					"type": "object",
					"properties": map[string]any{
						"answer": map[string]any{
							"type":        "string",
							"description": "the answer to the question, the answer should come from the text",
						},
					},
				},
			},
		},
		{
			Type: "function",
			Function: &llms.FunctionDefinition{
				Name:        "no_answer",
				Description: "Indicate that the text does not answer the question.",
				Parameters: map[string]any{
					"type": "object",
					"properties": map[string]any{
						"ignored": map[string]any{
							"type":        "string",
							"description": "ignored, just here to make sure the function is called.  Fill with anything.",
						},
					},
				},
			},
		},
	}

	answer := func(ctx context.Context, args string) (string, error) {
		type answer struct {
			Answer string `json:"answer"`
		}

		var a answer

		err := json.Unmarshal([]byte(args), &a)
		if err != nil {
			return "", err
		}

		return a.Answer, nil
	}

	noAnswer := func(ctx context.Context, ignored string) (string, error) {
		return "", nil
	}

	var req = []llms.MessageContent{
		{
			Role: llms.ChatMessageTypeSystem,
			Parts: []llms.ContentPart{
				llms.TextPart("Evaluate the given text to see if it answers the question from the user.  The text is as follows:"),
			},
		},
		{
			Role: llms.ChatMessageTypeSystem,
			Parts: []llms.ContentPart{
				llms.TextPart(text),
			},
		},
		{
			Role: llms.ChatMessageTypeHuman,
			Parts: []llms.ContentPart{
				llms.TextPart(q.Question),
			},
		},
	}

	res, err := q.Model.GenerateContent(ctx, req, llms.WithTools(availableTools))
	if err != nil {
		return "", err
	}

	if len(res.Choices) == 0 {
		return "", nil
	}

	if len(res.Choices[0].ToolCalls) == 0 {
		return "", nil
	}

	for _, call := range res.Choices[0].ToolCalls {
		switch call.FunctionCall.Name {
		case "answer":
			return answer(ctx, call.FunctionCall.Arguments)

		case "no_answer":
			return noAnswer(ctx, call.FunctionCall.Arguments)

		default:
			return "", fmt.Errorf("unknown function %s", call.FunctionCall.Name)
		}
	}

	return "", nil
}

func functionSearch(ctx context.Context, q Question, searchTerm string) (string, error) {

	slog.Info("searching", "search", searchTerm, "question", q)
	res, err := q.Search.Search(ctx, searchTerm)
	if err != nil {
		return "", err
	}

	if len(res) == 0 {
		return "", nil
	}

	// first pass try to see if any provide the result without needing archive
	for _, r := range res {
		trimmed := strings.TrimSpace(r.URL)
		if trimmed == "" {

		}

		slog.Info("extracting article", "url", trimmed)
		u, err := url.Parse(trimmed)
		if err != nil {
			continue
		}

		a, err := extractArticle(ctx, q.Cache, u)

		if err != nil {
			continue
		}

		slog.Info("extracted article", "url", a.URL, "title", a.Title, "body", a.Body)

		if a.Title != "" && a.Body != "" {
			answer, err := doesTextAnswerQuestion(ctx, q, a.Body)

			if err != nil {
				slog.Error("error checking if text answers question", "question", q.Question, "error", err)
				continue
			}

			if answer != "" {
				return answer, nil
			}
		}
	}

	return "", nil
}

func functionThink(ctx context.Context, q Question) (string, error) {
	fnAnswer := gollm.NewFunction(
		"answer",
		"Answer the question.",
		func(ctx context.Context, args struct {
			Answer string `description:"the answer to the question"`
		}) (string, error) {
			return args.Answer, nil
		})

	var temp float32 = 0.8
	req := gollm.Request{
		Messages: []gollm.Message{
			{
				Role: gollm.RoleSystem,
				Text: "Answer the given question as accurately and concisely as possible using the answer function.",
			},
			{
				Role: gollm.RoleUser,
				Text: q.Question,
			},
		},
		Toolbox:     gollm.NewToolBox(fnAnswer),
		Temperature: &temp,
	}

	res, err := q.Model.ChatComplete(ctx, req)

	if err != nil {
		return "", err
	}

	if len(res.Choices) == 0 {
		return "", nil
	}

	if len(res.Choices[0].Calls) == 0 {
		return "", nil
	}

	return req.Toolbox.Execute(ctx, res.Choices[0].Calls[0])
}

func appendResponse(req []llms.MessageContent, response *llms.ContentResponse) ([]llms.MessageContent, error) {
	if response == nil {
		return req, nil
	}

	if len(response.Choices) == 0 {
		return req, nil
	}

	choice := response.Choices[0]

	assistantResponse := llms.TextParts(llms.ChatMessageTypeAI, choice.Content)

	for _, tc := range choice.ToolCalls {
		assistantResponse.Parts = append(assistantResponse.Parts, tc)
	}

	return req, nil
}

func (o Options) Answer(ctx context.Context, q Question) (string, error) {
	a := agents.NewConversationalAgent()
}

func Answer(ctx context.Context, q Question) (string, error) {
	return DefaultOptions.Answer(ctx, q)
}
